/**
 * @file pixy2.c
 *
 * @author Jon McPhalen
 *
 * @copyright Copyright (c) Jon McPhalen, 2020. See end of file for
 * terms of use (MIT License). 
 *
 * @version 20 JAN 2020
 *
 * @brief Provides serial interface (up to 19.2K) to Pixy2 camera 
 * for color-connected functions (i.e., color tracking). This library
 * does not require a cog for serial communications.
 */


#include <stdint.h>
#include "pixy2.h"
#include "serial.h"

serial *pixy2;

uint8_t response[BUF_SIZE];


int32_t pixy2_start(uint8_t rxpin, uint8_t txpin, uint16_t baud)
{
  pixy2 = serial_open(rxpin, txpin, 0, baud);
  
  return (pixy2 != NULL) ? RESULT_OK : RESULT_ERROR;
}


int32_t pixy2_getVersion()
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_VERSION);
  serial_txChar(pixy2, 0x00);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_VERSION)
      return RESULT_OK;
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


uint16_t pixy2_hardware()
{
  if (response[2] == RESPONSE_VERSION)
    return extract16(6);
  else
    return 0;
}


uint8_t pixy2_fwMajor()
{
  if (response[2] == RESPONSE_VERSION)
    return response[8];
  else
    return 0;
}


uint8_t pixy2_fwMinor()
{
  if (response[2] == RESPONSE_VERSION)
    return response[9];
  else
    return 0;
}


uint16_t pixy2_fwBuild()
{
  if (response[2] == RESPONSE_VERSION)
    return extract16(10);
  else
    return 0;
}


char *pixy2_fwType()
{
  if (response[2] == RESPONSE_VERSION)
    return &response[12];
  else
    return NULL;
}


int32_t pixy2_getResolution(uint16_t *fWidth, uint16_t *fHeight)
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_RESOLUTION);
  serial_txChar(pixy2, 0x01);
  serial_txChar(pixy2, 0x00);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESOLUTION) {
      *fWidth  = extract16(6);
      *fHeight = extract16(8);
      return RESULT_OK;
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_setCameraBrightness(uint8_t brightness)
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_BRIGHTNESS);
  serial_txChar(pixy2, 0x01);
  serial_txChar(pixy2, brightness);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extract32(6);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_setServos(uint16_t s0, uint16_t s1)
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_SERVO);
  serial_txChar(pixy2, 0x04);
  serial_txChar(pixy2, s0 & 0xFF);
  serial_txChar(pixy2, s0 >> 8);
  serial_txChar(pixy2, s1 & 0xFF);
  serial_txChar(pixy2, s1 >> 8);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extract32(6);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_setLED(uint8_t r, uint8_t g, uint8_t b)
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_LED);
  serial_txChar(pixy2, 0x03);
  serial_txChar(pixy2, r);
  serial_txChar(pixy2, g);
  serial_txChar(pixy2, b);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extract32(6);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_setLamp(uint8_t upper, uint8_t lower)
{
  clearBuffer();

  if (upper)
    upper = 1;

  if (lower)
    lower = 1;

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_LAMP);
  serial_txChar(pixy2, 0x02);
  serial_txChar(pixy2, upper);
  serial_txChar(pixy2, lower);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extract32(6);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_getFPS()
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_FPS);
  serial_txChar(pixy2, 0x00);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extract32(6);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


uint8_t pixy2_getBlocks(uint8_t sigMap, uint8_t maxBlocks)
{
  clearBuffer();

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_BLOCKS);
  serial_txChar(pixy2, 0x02);
  serial_txChar(pixy2, sigMap);
  serial_txChar(pixy2, maxBlocks);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_BLOCKS) {
      return response[3] / 14;
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


int32_t pixy2_extractBlock(uint8_t i, uint16_t *block)
{
  uint8_t ofs = 6 + (i * BLOCK_SIZE);
  uint8_t j;

  if (response[2] == RESPONSE_BLOCKS) {
    if (i < (response[3] / 14)) {
      memcpy(block, &response[ofs], 14);
      return RESULT_OK;
    } 
  }     
  return RESULT_ERROR;
}


int32_t pixy2_getRGB(uint16_t x, uint16_t y, uint8_t saturate)
{
  clearBuffer();

  if (x > 315)
    x = 315;

  if (y > 207)
    y = 207;

  if (saturate > 1)
    saturate = 1;

  serial_txChar(pixy2, 0xAE);
  serial_txChar(pixy2, 0xC1);
  serial_txChar(pixy2, REQUEST_GET_RGB);
  serial_txChar(pixy2, 0x05);
  serial_txChar(pixy2, x & 0xFF);
  serial_txChar(pixy2, x >> 8);
  serial_txChar(pixy2, y & 0xFF);
  serial_txChar(pixy2, y >> 8);
  serial_txChar(pixy2, saturate);

  recvPacket();

  if (validateChecksum() != RESULT_OK) {
    return RESULT_CHECKSUM_ERROR;
  }
  else {
    if (response[2] == RESPONSE_RESULT) {
      return (int32_t)extractLE(6, 3);
    }
    else
      return (int32_t)(0xFFFFFF00 | response[2]);
  }
}


void clearBuffer()
{
  uint8_t i;

  for (i = 0; i < BUF_SIZE; i++)
    response[i] = 0;
}


uint8_t recvPacket()
{
  uint8_t len, b, payload;

  // get header
  for (len = 0; len < 6; len++) {
    b = serial_rxChar(pixy2);
    response[len] = (uint8_t)b;
  }

  // get size of payload
  payload = response[3];

  // get payload bytes if available
  while (payload-- > 0) {
    b = (uint8_t)serial_rxChar(pixy2);
    response[len++] = b;
  }

  return len;
}


int32_t validateChecksum()                                    // compare packet checksum to calcutated checksum
{
  uint16_t cs0 = extract16(4);
  uint16_t cs1 = 0;
  uint8_t  i   = response[3];                                 // length of payload
  uint8_t  j   = 6;                                           // starting index of payload bytes

  while (i-- > 0) {
    cs1 += response[j++];
  }

  return (cs1 == cs0) ? RESULT_OK : RESULT_ERROR;
}


uint16_t extract16(uint8_t i)
{
  return extractLE(i, 2);                                     // two bytes to 16-bit value
}


uint32_t extract32(uint8_t i)
{
  return extractLE(i, 4);                                     // four bytes to 32-bit value
}


uint32_t extractLE(uint8_t i, uint8_t n)                      // extract Little Endian value from response
{
  uint32_t result = 0;
  uint8_t j = 0;

  while (n-- > 0) {
    result |= response[i++] << j;
    j += 8;
  }
  return result;
}

// end of utilities



/**
 * TERMS OF USE: MIT License
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

